@@ -9,6 +9,12 @@ from gigagan_pytorch.gigagan_pytorch import (
 
 from gigagan_pytorch.unet_upsampler import UnetUpsampler
 
+from gigagan_pytorch.data import (
+    ImageDataset,
+    TextImageDataset,
+    MockTextImageDataset
+)
+
 __all__ = [
     GigaGAN,
     Generator,
@@ -16,5 +22,8 @@ __all__ = [
     AdaptiveConv2DMod,
     StyleNetwork,
     UnetUpsampler,
-    TextEncoder
-]
\ No newline at end of file
+    TextEncoder,
+    ImageDataset,
+    TextImageDataset,
+    MockTextImageDataset
+]
