@@ -2,11 +2,11 @@ from collections import namedtuple
 from pathlib import Path
 from math import log2, sqrt
 from functools import partial
-from random import randrange
+
+from torchvision import utils
 
 import torch
 import torch.nn.functional as F
-from torch.optim import Adam
 from torch import nn, einsum, Tensor
 from torch.autograd import grad as torch_grad
 from torch.utils.data import DataLoader
@@ -21,9 +21,12 @@ from ema_pytorch import EMA
 
 from gigagan_pytorch.version import __version__
 from gigagan_pytorch.open_clip import OpenClipAdapter
+from gigagan_pytorch.optimizer import get_optimizer
 
 from tqdm import tqdm
 
+from numerize import numerize
+
 # helpers
 
 def exists(val):
@@ -35,6 +38,9 @@ def default(*vals):
             return val
     return None
 
+def cast_tuple(t, length = 1):
+    return t if isinstance(t, tuple) else ((t,) * length)
+
 def is_power_of_two(n):
     return log2(n).is_integer()
 
@@ -54,6 +60,16 @@ def cycle(dl):
         for data in dl:
             yield data
 
+def num_to_groups(num, divisor):
+    groups, remainder = divmod(num, divisor)
+    arr = [divisor] * groups
+    if remainder > 0:
+        arr.append(remainder)
+    return arr
+
+def mkdir_if_not_exists(path):
+    path.mkdir(exist_ok = True, parents = True)
+
 # activation functions
 
 def leaky_relu(neg_slope = 0.1):
@@ -240,6 +256,17 @@ class AdaptiveConv2DMod(nn.Module):
 
         b, h = fmap.shape[0], fmap.shape[-2]
 
+        # account for feature map that has been expanded by the scale in the first dimension
+        # due to multiscale inputs and outputs
+
+        if mod.shape[0] != b:
+            mod = repeat(mod, 'b ... -> (s b) ...', s = b // mod.shape[0])
+
+        if kernel_mod.shape[0] != b:
+            kernel_mod = repeat(kernel_mod, 'b ... -> (s b) ...', s = b // kernel_mod.shape[0])
+
+        # prepare weights for modulation
+
         weights = self.weights
 
         if self.adaptive:
@@ -685,6 +712,25 @@ class StyleNetwork(nn.Module):
 
         return self.net(x)
 
+# noise
+
+class Noise(nn.Module):
+    def __init__(self, dim):
+        super().__init__()
+        self.weight = nn.Parameter(torch.zeros(dim, 1, 1))
+
+    def forward(
+        self,
+        x,
+        noise = None
+    ):
+        b, _, h, w, device = *x.shape, x.device
+
+        if not exists(noise):
+            noise = torch.randn(b, 1, h, w, device = device)
+
+        return x + self.weight * noise
+
 # generator
 
 class BaseGenerator(nn.Module):
@@ -695,10 +741,9 @@ class Generator(BaseGenerator):
     def __init__(
         self,
         *,
-        dim,
+        dim_capacity = 16,
         image_size,
         dim_max = 2048,
-        capacity = 16,
         channels = 3,
         style_network: Optional[Union[StyleNetwork, Dict]] = None,
         style_network_dim = None,
@@ -718,7 +763,6 @@ class Generator(BaseGenerator):
         unconditional = False
     ):
         super().__init__()
-        self.dim = dim
         self.channels = channels
 
         if isinstance(style_network, dict):
@@ -776,7 +820,7 @@ class Generator(BaseGenerator):
         resolutions = image_size / ((2 ** torch.arange(num_layers)))
         resolutions = resolutions.long().tolist()
 
-        dim_layers = (2 ** (torch.arange(num_layers) + 1)) * capacity
+        dim_layers = (2 ** (torch.arange(num_layers) + 1)) * dim_capacity
         dim_layers.clamp_(max = dim_max)
 
         dim_layers = torch.flip(dim_layers, (0,))
@@ -807,8 +851,10 @@ class Generator(BaseGenerator):
 
             resnet_block = nn.ModuleList([
                 adaptive_conv(dim_in, dim_out),
+                Noise(dim_out),
                 leaky_relu(),
                 adaptive_conv(dim_out, dim_out),
+                Noise(dim_out),
                 leaky_relu()
             ])
 
@@ -856,6 +902,10 @@ class Generator(BaseGenerator):
         if type(m) in {nn.Conv2d, nn.Linear}:
             nn.init.kaiming_normal_(m.weight, a = 0, mode = 'fan_in', nonlinearity = 'leaky_relu')
 
+    @property
+    def total_params(self):
+        return sum([p.numel() for p in self.parameters()])
+
     @property
     def device(self):
         return next(self.parameters()).device
@@ -920,7 +970,7 @@ class Generator(BaseGenerator):
 
         # main network
 
-        for squeeze_excite, (resnet_conv1, act1, resnet_conv2, act2), to_rgb_conv, self_attn, cross_attn, upsample, upsample_rgb in self.layers:
+        for squeeze_excite, (resnet_conv1, noise1, act1, resnet_conv2, noise2, act2), to_rgb_conv, self_attn, cross_attn, upsample, upsample_rgb in self.layers:
 
             if exists(squeeze_excite):
                 skip_excite = squeeze_excite(x)
@@ -931,8 +981,11 @@ class Generator(BaseGenerator):
                 x = x * excite
 
             x = resnet_conv1(x, mod = next(conv_mods), kernel_mod = next(conv_mods))
+            x = noise1(x)
             x = act1(x)
+
             x = resnet_conv2(x, mod = next(conv_mods), kernel_mod = next(conv_mods))
+            x = noise2(x)
             x = act2(x)
 
             if exists(self_attn):
@@ -967,10 +1020,17 @@ class SimpleDecoder(nn.Module):
         dim,
         *,
         dims: Tuple[int, ...],
-        num_patches = 1
+        patch_dim: int = 1,
+        frac_patches: float = 1.,
+        dropout: float = 0.5
     ):
         super().__init__()
-        self.num_patches = num_patches
+        assert 0 < frac_patches <= 1.
+
+        self.patch_dim = patch_dim
+        self.frac_patches = frac_patches
+
+        self.dropout = nn.Dropout(dropout)
 
         dims = [dim, *dims]
 
@@ -979,27 +1039,41 @@ class SimpleDecoder(nn.Module):
         for dim_in, dim_out in zip(dims[:-1], dims[1:]):
             layers.append(nn.Sequential(
                 Upsample(dim_in),
-                conv2d_3x3(dim_in, dim_out * 2),
-                nn.GLU(dim = 1)
+                conv2d_3x3(dim_in, dim_out),
+                leaky_relu()
             ))
 
         self.net = nn.Sequential(*layers)
 
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
     def forward(
         self,
         fmap,
         orig_image
     ):
-        if self.num_patches > 1:
+        fmap = self.dropout(fmap)
+
+        if self.frac_patches < 1.:
+            batch, patch_dim = fmap.shape[0], self.patch_dim
             fmap_size, img_size = fmap.shape[-1], orig_image.shape[-1]
 
-            assert divisible_by(fmap_size, self.num_patches)
-            assert divisible_by(img_size, self.num_patches)
+            assert divisible_by(fmap_size, patch_dim), f'feature map dimensions are {fmap_size}, but the patch dim was designated to be {patch_dim}'
+            assert divisible_by(img_size, patch_dim), f'image size is {img_size} but the patch dim was specified to be {patch_dim}'
 
-            fmap, orig_image = map(lambda t: rearrange(t, 'b c (p1 h) (p2 w) -> (p1 p2) b c h w', p1 = self.num_patches, p2 = self.num_patches), (fmap, orig_image))
+            fmap, orig_image = map(lambda t: rearrange(t, 'b c (p1 h) (p2 w) -> b (p1 p2) c h w', p1 = patch_dim, p2 = patch_dim), (fmap, orig_image))
 
-            rand_index = randrange(0, self.num_patches ** 2)
-            fmap, orig_image = map(lambda t: t[rand_index], (fmap, orig_image))
+            total_patches = patch_dim ** 2
+            num_patches_recon = max(int(self.frac_patches * total_patches), 1)
+
+            batch_arange = torch.arange(batch, device = self.device)[..., None]
+            batch_randperm = torch.randn((batch, total_patches)).sort(dim = -1).indices
+            patch_indices = batch_randperm[..., :num_patches_recon]
+
+            fmap, orig_image = map(lambda t: t[batch_arange, patch_indices], (fmap, orig_image))
+            fmap, orig_image = map(lambda t: rearrange(t, 'b p ... -> (b p) ...'), (fmap, orig_image))
 
         recon = self.net(fmap)
         return F.mse_loss(recon, orig_image)
@@ -1125,7 +1199,12 @@ class Predictor(nn.Module):
         klass = nn.Conv2d if unconditional else partial(AdaptiveConv2DMod, num_conv_kernels = num_conv_kernels)
 
         for ind in range(depth):
-            self.layers.append(klass(dim, dim, 1))
+            self.layers.append(nn.ModuleList([
+                klass(dim, dim, 3, padding = 1),
+                leaky_relu(),
+                klass(dim, dim, 3, padding = 1),
+                leaky_relu()
+            ]))
 
         self.to_logits = nn.Conv2d(dim, 1, 1)
 
@@ -1137,12 +1216,21 @@ class Predictor(nn.Module):
     ):
         residual = self.residual_fn(x)
 
-        for layer in self.layers:
-            kwargs = dict()
-            if not self.unconditional:
-                kwargs = dict(mod = mod, kernel_mod = kernel_mod)
+        kwargs = dict()
 
-            x = layer(x, **kwargs)
+        if not self.unconditional:
+            kwargs = dict(mod = mod, kernel_mod = kernel_mod)
+
+        for conv1, activation, conv2, activation in self.layers:
+
+            inner_residual = x
+
+            x = conv1(x, **kwargs)
+            x = activation(x)
+            x = conv2(x, **kwargs)
+            x = activation(x)
+
+            x = x + inner_residual
 
         x = x + residual
         return self.to_logits(x)
@@ -1152,9 +1240,8 @@ class Discriminator(nn.Module):
     def __init__(
         self,
         *,
-        dim,
+        dim_capacity = 16,
         image_size,
-        capacity = 16,
         dim_max = 2048,
         channels = 3,
         attn_resolutions: Tuple[int, ...] = (32, 16),
@@ -1164,15 +1251,18 @@ class Discriminator(nn.Module):
         ff_mult = 4,
         text_encoder: Optional[Union[TextEncoder, Dict]] = None,
         text_dim = None,
+        filter_input_resolutions: bool = True,
         multiscale_input_resolutions: Tuple[int, ...] = (64, 32, 16, 8),
-        multiscale_output_resolutions: Tuple[int, ...] = (32, 16, 8, 4),
+        multiscale_output_skip_stages: int = 1,
         aux_recon_resolutions: Tuple[int, ...] = (8,),
-        aux_recon_patches: Tuple[int, ...] = (2,),
+        aux_recon_patch_dims: Tuple[int, ...] = (2,),
+        aux_recon_frac_patches: Tuple[float, ...] = (0.25,),
+        aux_recon_fmap_dropout: float = 0.5,
         resize_mode = 'bilinear',
         num_conv_kernels = 2,
         num_skip_layers_excite = 0,
         unconditional = False,
-        scale_invariant_training = True
+        predictor_depth = 2
     ):
         super().__init__()
         self.unconditional = unconditional
@@ -1181,16 +1271,20 @@ class Discriminator(nn.Module):
         assert is_power_of_two(image_size)
         assert all([*map(is_power_of_two, attn_resolutions)])
 
+        if filter_input_resolutions:
+            multiscale_input_resolutions = [*filter(lambda t: t < image_size, multiscale_input_resolutions)]
+
         assert is_unique(multiscale_input_resolutions)
         assert all([*map(is_power_of_two, multiscale_input_resolutions)])
-        assert all([*map(lambda t: t >= 4, multiscale_input_resolutions)])
         assert all([*map(lambda t: t < image_size, multiscale_input_resolutions)])
 
         self.multiscale_input_resolutions = multiscale_input_resolutions
 
-        assert is_unique(multiscale_output_resolutions)
-        assert all([*map(is_power_of_two, multiscale_output_resolutions)])
+        assert multiscale_output_skip_stages > 0
+        multiscale_output_resolutions = [resolution // (2 ** multiscale_output_skip_stages) for resolution in multiscale_input_resolutions]
+
         assert all([*map(lambda t: t >= 4, multiscale_output_resolutions)])
+
         assert all([*map(lambda t: t < image_size, multiscale_output_resolutions)])
 
         if len(multiscale_input_resolutions) > 0 and len(multiscale_output_resolutions) > 0:
@@ -1200,8 +1294,9 @@ class Discriminator(nn.Module):
         self.multiscale_output_resolutions = multiscale_output_resolutions
 
         assert all([*map(is_power_of_two, aux_recon_resolutions)])
-        assert len(aux_recon_resolutions) == len(aux_recon_patches)
-        self.aux_recon_resolutions_to_patches = {resolution: patches for resolution, patches in zip(aux_recon_resolutions, aux_recon_patches)}
+        assert len(aux_recon_resolutions) == len(aux_recon_patch_dims) == len(aux_recon_frac_patches)
+
+        self.aux_recon_resolutions_to_patches = {resolution: (patch_dim, frac_patches) for resolution, patch_dim, frac_patches in zip(aux_recon_resolutions, aux_recon_patch_dims, aux_recon_frac_patches)}
 
         self.resize_mode = resize_mode
 
@@ -1212,7 +1307,7 @@ class Discriminator(nn.Module):
         resolutions = image_size / ((2 ** torch.arange(num_layers)))
         resolutions = resolutions.long().tolist()
 
-        dim_layers = (2 ** (torch.arange(num_layers) + 1)) * capacity
+        dim_layers = (2 ** (torch.arange(num_layers) + 1)) * dim_capacity
         dim_layers = F.pad(dim_layers, (1, 0), value = channels)
         dim_layers.clamp_(max = dim_max)
 
@@ -1236,7 +1331,6 @@ class Discriminator(nn.Module):
             should_skip_layer_excite = not is_first and num_skip_layers_excite > 0 and (ind + num_skip_layers_excite) < len(dim_pairs)
 
             has_attn = resolution in attn_resolutions
-            has_multiscale_input = resolution in multiscale_input_resolutions
             has_multiscale_output = resolution in multiscale_output_resolutions
 
             has_aux_recon_decoder = resolution in aux_recon_resolutions
@@ -1249,13 +1343,7 @@ class Discriminator(nn.Module):
 
             # multi-scale rgb input to feature dimension
 
-            from_rgb = None
-            if has_multiscale_input and scale_invariant_training:
-                from_rgb = nn.Conv2d(channels, dim_in, 7, padding = 3)
-
-            # dim in + channels for the main features
-
-            dim_in = dim_in + (channels if has_multiscale_input else 0)
+            from_rgb = nn.Conv2d(channels, dim_in, 7, padding = 3)
 
             # residual convolution
 
@@ -1275,18 +1363,20 @@ class Discriminator(nn.Module):
             multiscale_output_predictor = None
 
             if has_multiscale_output:
-                multiscale_output_predictor = Predictor(dim_out, num_conv_kernels = num_conv_kernels, unconditional = unconditional)
+                multiscale_output_predictor = Predictor(dim_out, num_conv_kernels = num_conv_kernels, depth = 2, unconditional = unconditional)
                 predictor_dims.extend([dim_out, dim_kernel_attn])
 
             aux_recon_decoder = None
 
             if has_aux_recon_decoder:
-                num_patches = self.aux_recon_resolutions_to_patches[resolution]
+                patch_dim, frac_patches = self.aux_recon_resolutions_to_patches[resolution]
 
                 aux_recon_decoder = SimpleDecoder(
                     dim_out,
                     dims = tuple(upsample_dims),
-                    num_patches = num_patches
+                    patch_dim = patch_dim,
+                    frac_patches = frac_patches,
+                    dropout = aux_recon_fmap_dropout
                 )
 
             self.layers.append(nn.ModuleList([
@@ -1332,6 +1422,14 @@ class Discriminator(nn.Module):
     def resize_image_to(self, images, resolution):
         return F.interpolate(images, resolution, mode = self.resize_mode)
 
+    @property
+    def total_params(self):
+        return sum([p.numel() for p in self.parameters()])
+
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
     @beartype
     def forward(
         self,
@@ -1340,8 +1438,9 @@ class Discriminator(nn.Module):
         texts: Optional[List[str]] = None,
         text_embeds = None,
         real_images = None,                   # if this were passed in, the network will automatically append the real to the presumably generated images passed in as the first argument, and generate all intermediate resolutions through resizing and concat appropriately
-        return_all_aux_loss = False           # this would return auxiliary reconstruction loss for both fake and real
-
+        return_all_aux_loss = False,          # this would return auxiliary reconstruction loss for both fake and real
+        return_multiscale_outputs = True,     # can force it not to return multi-scale logits
+        calc_aux_loss = True
     ):
         if not self.unconditional:
             assert exists(texts) ^ exists(text_embeds)
@@ -1360,14 +1459,20 @@ class Discriminator(nn.Module):
 
         x = images
 
-        assert x.shape[-2:] == (self.image_size, self.image_size)
+        image_size = (self.image_size, self.image_size)
+        assert x.shape[-2:] == image_size
 
         # if real images are passed in, assume `images` are generated, and take care of all the multi-resolution input. this can also be done externally, in which case `real_images` will not be populated
 
         has_real_images = exists(real_images)
 
         if has_real_images:
-            split_batch_size = (x.shape[0], real_images.shape[0])
+            assert real_images.shape[-2:] == image_size, f'images from dataloader must be of size {image_size} but received something else'
+
+            num_fake_images = x.shape[0]
+            num_real_images = real_images.shape[0]
+
+            split_batch_size = (num_fake_images, num_real_images)
             x = torch.cat((x, real_images), dim = 0)
 
         batch = x.shape[0]
@@ -1403,11 +1508,15 @@ class Discriminator(nn.Module):
                 excitations.append(skip_excite)
 
             excite = safe_unshift(excitations)
+
             if exists(excite):
                 excite = repeat(excite, 'b ... -> (s b) ...', s = x.shape[0] // excite.shape[0])
                 x = x * excite
 
-            if resolution in self.multiscale_input_resolutions:
+            batch_prev_stage = x.shape[0]
+            has_multiscale_input = resolution in self.multiscale_input_resolutions
+
+            if has_multiscale_input:
                 images_to_concat = rgbs_index.get(resolution, None)
 
                 # if no rgbs passed in, assume all real images, and just resize, though realistically you would concat fake and real images together using helper function `create_real_fake_rgbs` function
@@ -1415,18 +1524,21 @@ class Discriminator(nn.Module):
                 if not exists(images_to_concat):
                     images_to_concat = self.resize_image_to(images, resolution)
 
-                images_to_concat = repeat(images_to_concat, 'b ... -> (s b) ...', s = x.shape[0] // images_to_concat.shape[0])
+                # multi-scale input features
 
-                # concat the rgb (or real images reshaped)
+                multi_scale_input_feats = from_rgb(images_to_concat)
 
-                x = torch.cat((images_to_concat, x), dim = 1)
+                # expand multi-scale input features, as could include extra scales from previous stage
 
-                # concat the rgb, projected into the feature dimension space
+                multi_scale_input_feats = repeat(multi_scale_input_feats, 'b ... -> (s b) ...', s = x.shape[0] // images_to_concat.shape[0])
 
-                if exists(from_rgb):
-                    multi_scale_x = from_rgb(images_to_concat)
-                    multi_scale_x = torch.cat((images_to_concat, multi_scale_x), dim = 1)
-                    x = torch.cat((x, multi_scale_x), dim = 0)
+                # add the multi-scale input features to the current hidden state from main stem
+
+                images_to_concat = x + multi_scale_input_feats
+
+                # and also concat for scale invariance
+
+                x = torch.cat((x, multi_scale_input_feats), dim = 0)
 
             residual = residual_fn(x)
             x = block(x)
@@ -1439,7 +1551,9 @@ class Discriminator(nn.Module):
                 if not self.unconditional:
                     pred_kwargs = dict(mod = next(conv_mods), kernel_mod = next(conv_mods))
 
-                multiscale_outputs.append(predictor(x, **pred_kwargs))
+                if return_multiscale_outputs:
+                    predictor_input = x[:batch_prev_stage]
+                    multiscale_outputs.append(predictor(predictor_input, **pred_kwargs))
 
             if exists(downsample):
                 x = downsample(x)
@@ -1447,17 +1561,23 @@ class Discriminator(nn.Module):
             x = x + residual
             x = x * self.residual_scale
 
-            if exists(recon_decoder) and (return_all_aux_loss or has_real_images):
+            if exists(recon_decoder) and calc_aux_loss and (return_all_aux_loss or has_real_images):
 
                 if return_all_aux_loss:
-                    recon_output = x
+                    recon_output = x[:batch_prev_stage]
 
-                elif has_real_images:
-                    recon_output = rearrange(x, '(s b) ... -> s b ...', b = batch)
+                recon_output = rearrange(x, '(s b) ... -> s b ...', b = batch)
+
+                if has_real_images:
                     _, recon_output = recon_output.split(split_batch_size, dim = 1)
-                    recon_output = rearrange(recon_output, 's b ... -> (s b) ...')
 
-                aux_recon_target = repeat(aux_recon_target, 'b ... -> (s b) ...', s = recon_output.shape[0] // aux_recon_target.shape[0])
+                # only use the input real images for aux recon
+
+                recon_output = recon_output[0]
+
+                # only reconstruct a fraction of images across batch and scale
+                # for efficiency
+
                 aux_recon_loss = recon_decoder(recon_output, aux_recon_target)
                 aux_recon_losses.append(aux_recon_loss)
 
@@ -1502,22 +1622,34 @@ class GigaGAN(nn.Module):
         generator: Union[BaseGenerator, Dict],
         discriminator: Union[Discriminator, Dict],
         text_encoder: Optional[Union[TextEncoder, Dict]] = None,
-        learning_rate = 1e-4,
-        betas = (0.9, 0.99),
+        learning_rate = 2e-4,
+        betas = (0.5, 0.9),
+        weight_decay = 0.,
         discr_aux_recon_loss_weight = 0.25,
-        multiscale_divergence_loss_weight = 1.,
-        apply_gradient_penalty_every = 16,
+        multiscale_divergence_loss_weight = 0.1,
+        calc_multiscale_loss_every = 1,
+        apply_gradient_penalty_every = 4,
+        ttur_mult = 1,
         train_upsampler = False,
         upsampler_replace_rgb_with_input_lowres_image = False,
         log_steps_every = 20,
-        create_ema_generator_at_init = True
+        create_ema_generator_at_init = True,
+        save_and_sample_every = 1000,
+        early_save_thres_steps = 2500,
+        early_save_and_sample_every = 100,
+        num_samples = 25,
+        model_folder = './gigagan-models',
+        results_folder = './gigagan-results',
+        val_upsampler_dl: DataLoader = None
     ):
         super().__init__()
 
         self.train_upsampler = train_upsampler
 
-        self.upsampler_replace_rgb_with_input_lowres_image= upsampler_replace_rgb_with_input_lowres_image
+        self.upsampler_replace_rgb_with_input_lowres_image = upsampler_replace_rgb_with_input_lowres_image
+
         self.apply_gradient_penalty_every = apply_gradient_penalty_every
+        self.calc_multiscale_loss_every = calc_multiscale_loss_every
 
         if train_upsampler:
             from gigagan_pytorch.unet_upsampler import UnetUpsampler
@@ -1536,6 +1668,11 @@ class GigaGAN(nn.Module):
         self.G = generator
         self.D = discriminator
 
+        # print number of parameters
+
+        print(f'Generator parameters: {numerize.numerize(generator.total_params)}')
+        print(f'Discriminator parameters: {numerize.numerize(discriminator.total_params)}')
+
         # text encoder
 
         if exists(text_encoder):
@@ -1550,8 +1687,8 @@ class GigaGAN(nn.Module):
 
         # optimizers
 
-        self.G_opt = Adam(self.G.parameters(), lr = learning_rate, betas = betas)
-        self.D_opt = Adam(self.D.parameters(), lr = learning_rate, betas = betas)
+        self.G_opt = get_optimizer(self.G.parameters(), lr = learning_rate, betas = betas, weight_decay = weight_decay)
+        self.D_opt = get_optimizer(self.D.parameters(), lr = learning_rate * ttur_mult, betas = betas, weight_decay = weight_decay)
 
         # loss related
 
@@ -1571,9 +1708,25 @@ class GigaGAN(nn.Module):
 
         self.register_buffer('steps', torch.ones(1, dtype = torch.long))
 
+        # save and sample
+
+        self.save_and_sample_every = save_and_sample_every
+        self.early_save_thres_steps = early_save_thres_steps
+        self.early_save_and_sample_every = early_save_and_sample_every
+
+        self.num_samples = num_samples
+
+        self.val_dl_iter = val_upsampler_dl
+
+        self.results_folder = Path(results_folder)
+        self.model_folder = Path(model_folder)
+
+        mkdir_if_not_exists(self.results_folder)
+        mkdir_if_not_exists(self.model_folder)
+
     def save(self, path, overwrite = True):
         path = Path(path)
-        path.parents[0].mkdir(parents = True, exist_ok =True)
+        mkdir_if_not_exists(path.parents[0])
 
         assert overwrite or not path.exists()
 
@@ -1637,17 +1790,64 @@ class GigaGAN(nn.Module):
     def print(self, msg):
         print(msg)
 
+    def generate_kwargs(self, dl_iter, batch_size):
+        # what to pass into the generator
+        # depends on whether training upsampler or not
+
+        maybe_text_kwargs = dict()
+        if self.train_upsampler or not self.unconditional:
+            assert exists(dl_iter)
+
+            if self.unconditional:
+                real_images = next(dl_iter)
+            else:
+                result = next(dl_iter)
+                assert isinstance(result, tuple), 'dataset should return a tuple of two items for text conditioned training, (images: Tensor, texts: List[str])'
+                real_images, texts = result
+
+                maybe_text_kwargs['texts'] = texts
+
+            real_images = real_images.to(self.device)
+
+        # if training upsample generator, need to downsample real images
+
+        if self.train_upsampler:
+            size = self.G.input_image_size
+            lowres_real_images = F.interpolate(real_images, (size, size))
+
+            G_kwargs = dict(
+                lowres_image = lowres_real_images,
+                replace_rgb_with_input_lowres_image = self.upsampler_replace_rgb_with_input_lowres_image
+            )
+        else:
+            assert exists(batch_size)
+
+            G_kwargs = dict(batch_size = batch_size)
+
+        # create noise
+
+        noise = torch.randn(batch_size, self.G.style_network.dim, device = self.device)
+
+        G_kwargs.update(noise = noise)
+
+        return G_kwargs, maybe_text_kwargs
+    
     @beartype
     def train_discriminator_step(
         self,
         dl_iter: Iterable,
         grad_accum_every = 1,
-        apply_gradient_penalty = False
+        apply_gradient_penalty = False,
+        calc_multiscale_loss = True
     ):
         total_divergence = 0.
         total_gp_loss = 0.
         total_aux_loss = 0.
-        total_multiscale_divergence = 0.
+
+        total_multiscale_divergence = 0. if calc_multiscale_loss else None
+
+        self.G.train()
+        self.D.train()
 
         self.D_opt.zero_grad()
 
@@ -1669,23 +1869,7 @@ class GigaGAN(nn.Module):
 
             # for discriminator training, fit upsampler and image synthesis logic under same function
 
-            if self.train_upsampler:
-                size = self.G.input_image_size
-                lowres_real_images = F.interpolate(real_images, (size, size))
-
-                G_kwargs = dict(
-                    lowres_image = lowres_real_images,
-                    replace_rgb_with_input_lowres_image = self.upsampler_replace_rgb_with_input_lowres_image
-                )
-            else:
-                G_kwargs = dict(batch_size = batch_size)
-
-            # add texts for conditioning if needed
-
-            maybe_text_kwargs = dict()
-
-            if not self.unconditional:
-                maybe_text_kwargs['texts'] = texts
+            G_kwargs, maybe_text_kwargs = self.generate_kwargs(dl_iter, batch_size)
 
             # generator
 
@@ -1708,12 +1892,13 @@ class GigaGAN(nn.Module):
             (fake_logits, real_logits), multiscale_logits, aux_recon_losses = self.D(
                 images,
                 rgbs,
-                real_images = real_images,
                 **maybe_text_kwargs,
+                real_images = real_images,
+                return_multiscale_outputs = calc_multiscale_loss
             )
 
             divergence = discriminator_hinge_loss(real_logits, fake_logits)
-            total_divergence = total_divergence + (divergence / grad_accum_every)
+            total_divergence += (divergence.item() / grad_accum_every)
 
             # handle multi-scale divergence
 
@@ -1721,14 +1906,14 @@ class GigaGAN(nn.Module):
 
             multiscale_real_logits = []
 
-            if self.multiscale_divergence_loss_weight > 0.:
+            if self.multiscale_divergence_loss_weight > 0. and len(multiscale_logits) > 0:
 
                 for multiscale_fake, multiscale_real in multiscale_logits:
-                    multiscale_loss = discriminator_hinge_loss(multiscale_fake, multiscale_real)
+                    multiscale_loss = discriminator_hinge_loss(multiscale_real, multiscale_fake)
                     multiscale_divergence = multiscale_divergence + multiscale_loss
                     multiscale_real_logits.append(multiscale_real)
 
-                total_multiscale_divergence += multiscale_divergence / grad_accum_every
+                total_multiscale_divergence += (multiscale_divergence.item() / grad_accum_every)
 
             # figure out gradient penalty if needed
 
@@ -1741,7 +1926,7 @@ class GigaGAN(nn.Module):
                     grad_output_weights = [1., *(self.multiscale_divergence_loss_weight,) * len(multiscale_real_logits)]
                 )
 
-                total_gp_loss += gp_loss / grad_accum_every
+                total_gp_loss += (gp_loss.item() / grad_accum_every)
 
             # sum up losses
 
@@ -1753,7 +1938,7 @@ class GigaGAN(nn.Module):
             if self.discr_aux_recon_loss_weight > 0.:
                 aux_loss = sum(aux_recon_losses)
 
-                total_aux_loss += aux_loss / grad_accum_every
+                total_aux_loss += (aux_loss.item() / grad_accum_every)
 
                 total_loss = total_loss + aux_loss * self.discr_aux_recon_loss_weight
 
@@ -1769,51 +1954,23 @@ class GigaGAN(nn.Module):
         self,
         batch_size = None,
         dl_iter: Optional[Iterable] = None,
-        grad_accum_every = 1
+        grad_accum_every = 1,
+        calc_multiscale_loss = True
     ):
         total_divergence = 0.
-        total_multiscale_divergence = 0.
+        total_multiscale_divergence = 0. if calc_multiscale_loss else None
+
+        self.G.train()
+        self.D.train()
 
         self.D_opt.zero_grad()
         self.G_opt.zero_grad()
 
         for _ in range(grad_accum_every):
 
-            # what to pass into the generator
-            # depends on whether training upsampler or not
-
-            maybe_text_kwargs = dict()
-
-            if self.train_upsampler or not self.unconditional:
-                assert exists(dl_iter)
-
-                if self.unconditional:
-                    real_images = next(dl_iter)
-                else:
-                    result = next(dl_iter)
-                    assert isinstance(result, tuple), 'dataset should return a tuple of two items for text conditioned training, (images: Tensor, texts: List[str])'
-                    real_images, texts = result
-
-                    maybe_text_kwargs['texts'] = texts
-
-                real_images = real_images.to(self.device)
-
-            # if training upsample generator, need to downsample real images
-
-            if self.train_upsampler:
-                size = self.G.input_image_size
-                lowres_real_images = F.interpolate(real_images, (size, size))
-
-                G_kwargs = dict(
-                    lowres_image = lowres_real_images,
-                    replace_rgb_with_input_lowres_image = self.upsampler_replace_rgb_with_input_lowres_image
-                )
-            else:
-                assert exists(batch_size)
-
-                G_kwargs = dict(batch_size = batch_size)
-
             # generator
+            
+            G_kwargs, maybe_text_kwargs = self.generate_kwargs(dl_iter, batch_size)
 
             image, rgbs = self.G(
                 **G_kwargs,
@@ -1826,24 +1983,26 @@ class GigaGAN(nn.Module):
             logits, multiscale_logits, _ = self.D(
                 image,
                 rgbs,
-                **maybe_text_kwargs
+                **maybe_text_kwargs,
+                return_multiscale_outputs = calc_multiscale_loss,
+                calc_aux_loss = False
             )
 
             # hinge loss
 
             divergence = generator_hinge_loss(logits)
 
-            total_divergence += divergence
+            total_divergence += (divergence.item() / grad_accum_every)
 
             total_loss = divergence
 
-            if self.multiscale_divergence_loss_weight > 0.:
+            if self.multiscale_divergence_loss_weight > 0. and len(multiscale_logits) > 0:
                 multiscale_divergence = 0.
 
                 for multiscale_logit in multiscale_logits:
                     multiscale_divergence = multiscale_divergence + generator_hinge_loss(multiscale_logit)
 
-                total_multiscale_divergence += multiscale_divergence
+                total_multiscale_divergence += (multiscale_divergence.item() / grad_accum_every)
 
                 total_loss = total_loss + multiscale_divergence * self.multiscale_divergence_loss_weight
 
@@ -1858,6 +2017,56 @@ class GigaGAN(nn.Module):
 
         return TrainGenLosses(total_divergence, total_multiscale_divergence)
 
+    def sample_lambda(self, dl_iter, batch_size):
+        G_kwargs, maybe_text_kwargs = self.generate_kwargs(dl_iter, batch_size)
+
+        generator_output = self.G(**G_kwargs, **maybe_text_kwargs)
+
+        if not self.train_upsampler:
+            return generator_output
+
+        output_size = generator_output.shape[-1]
+        lowres_image = G_kwargs['lowres_image']
+        lowres_image = F.interpolate(lowres_image, (output_size, output_size))
+
+        return torch.cat([lowres_image, generator_output])
+
+    @torch.inference_mode()
+    def save_sample(
+        self,
+        batch_size,
+        dl_iter = None
+    ):
+        milestone = self.steps.item() // self.save_and_sample_every
+        nrow_mult = 2 if self.train_upsampler else 1
+        batches = num_to_groups(self.num_samples, batch_size)
+
+        dl_iter = default(self.val_dl_iter, dl_iter)
+
+        assert exists(dl_iter)
+
+        sample_models_and_output_file_name = [(self.G, f'sample-{milestone}.png')]
+
+        if self.has_ema_generator:
+            sample_models_and_output_file_name.append((self.G_ema, f'ema-sample-{milestone}.png'))
+
+        for model, filename in sample_models_and_output_file_name:
+            model.eval()
+
+            all_images_list = list(map(lambda n: self.sample_lambda(dl_iter, n), batches))
+            all_images = torch.cat(all_images_list, dim=0)
+
+            all_images.clamp_(0., 1.)
+
+            utils.save_image(
+                all_images,
+                str(self.results_folder / filename),
+                nrow = int(sqrt(self.num_samples)) * nrow_mult
+            )
+
+        # Possible to do: Include some metric to save if improved, include some sampler dict text entries
+        self.save(str(self.model_folder / f'model-{milestone}.ckpt'))
+        
     @beartype
     def forward(
         self,
@@ -1870,22 +2079,45 @@ class GigaGAN(nn.Module):
         dl_iter = cycle(dataloader)
 
         last_gp_loss = 0.
+        last_multiscale_d_loss = 0.
+        last_multiscale_g_loss = 0.
 
         for _ in tqdm(range(steps), initial = self.steps.item()):
             steps = self.steps.item()
             is_first_step = steps == 1
 
             apply_gradient_penalty = self.apply_gradient_penalty_every > 0 and divisible_by(steps, self.apply_gradient_penalty_every)
+            calc_multiscale_loss =  self.calc_multiscale_loss_every > 0 and divisible_by(steps, self.calc_multiscale_loss_every)
 
-            d_loss, multiscale_d_loss, gp_loss, recon_loss = self.train_discriminator_step(dl_iter, grad_accum_every = grad_accum_every, apply_gradient_penalty = apply_gradient_penalty)
-            g_loss, multiscale_g_loss = self.train_generator_step(dl_iter = dl_iter, batch_size = batch_size, grad_accum_every = grad_accum_every)
+            d_loss, multiscale_d_loss, gp_loss, recon_loss = self.train_discriminator_step(
+                dl_iter,
+                grad_accum_every = grad_accum_every,
+                apply_gradient_penalty = apply_gradient_penalty,
+                calc_multiscale_loss = calc_multiscale_loss
+            )
+
+            g_loss, multiscale_g_loss = self.train_generator_step(
+                dl_iter = dl_iter,
+                batch_size = batch_size,
+                grad_accum_every = grad_accum_every,
+                calc_multiscale_loss = calc_multiscale_loss
+            )
 
             if exists(gp_loss):
                 last_gp_loss = gp_loss
 
-            if is_first_step or divisible_by(steps, self.log_steps_every):
-                self.print(f' G: {g_loss:.2f} | MSG: {multiscale_g_loss:.2f} | D: {d_loss:.2f} | MSD: {multiscale_d_loss:.2f} | GP: {last_gp_loss:.2f} | SSL: {recon_loss:.2f}')
+            if exists(multiscale_d_loss):
+                last_multiscale_d_loss = multiscale_d_loss
 
+            if exists(multiscale_g_loss):
+                last_multiscale_g_loss = multiscale_g_loss
+
+            if is_first_step or divisible_by(steps, self.log_steps_every):
+                self.print(f' G: {g_loss:.2f} | MSG: {last_multiscale_g_loss:.2f} | D: {d_loss:.2f} | MSD: {last_multiscale_d_loss:.2f} | GP: {last_gp_loss:.2f} | SSL: {recon_loss:.2f}')
+
+            if is_first_step or divisible_by(steps, self.save_and_sample_every) or (steps <= self.early_save_thres_steps and divisible_by(steps, self.early_save_and_sample_every)):
+                self.save_sample(batch_size, dl_iter)
+            
             self.steps += 1
 
         self.print(f'complete {steps} training steps')
